local groups = {}

local function get_base_colors(colors, contrast)
  local bg = vim.o.background
  local base_colors = {
    -- options (dark mode by default)
    bg0 = colors.dark0,
    bg1 = colors.dark1,
    bg2 = colors.dark2,
    bg3 = colors.dark3,
    bg4 = colors.dark4,
    fg0 = colors.light0,
    fg1 = colors.light1,
    fg2 = colors.light2,
    fg3 = colors.light3,
    fg4 = colors.light4,
    red = colors.bright_red,
    green = colors.bright_green,
    yellow = colors.bright_yellow,
    blue = colors.bright_blue,
    purple = colors.bright_purple,
    aqua = colors.bright_aqua,
    orange = colors.bright_orange,
    neutral_red = colors.neutral_red,
    neutral_green = colors.neutral_green,
    neutral_yellow = colors.neutral_yellow,
    neutral_blue = colors.neutral_blue,
    neutral_purple = colors.neutral_purple,
    neutral_aqua = colors.neutral_aqua,
  }

  local light_colors = {
    bg0 = colors.light0,
    bg1 = colors.light1,
    bg2 = colors.light2,
    bg3 = colors.light3,
    bg4 = colors.light4,
    fg0 = colors.dark0,
    fg1 = colors.dark1,
    fg2 = colors.dark2,
    fg3 = colors.dark3,
    fg4 = colors.dark4,
    red = colors.faded_red,
    green = colors.faded_green,
    yellow = colors.faded_yellow,
    blue = colors.faded_blue,
    purple = colors.faded_purple,
    aqua = colors.faded_aqua,
    orange = colors.faded_orange,
    neutral_red = colors.neutral_red,
    neutral_green = colors.neutral_green,
    neutral_yellow = colors.neutral_yellow,
    neutral_blue = colors.neutral_blue,
    neutral_purple = colors.neutral_purple,
    neutral_aqua = colors.neutral_aqua,
  }

  -- swap colors if light mode
  if bg == "light" then
    base_colors = light_colors
  end

  if contrast == "hard" then
    base_colors.bg0 = colors[bg .. "0_hard"]
  elseif contrast == "soft" then
    base_colors.bg0 = colors[bg .. "0_soft"]
  end

  base_colors.gray = colors.gray

  return base_colors
end

-- neovim terminal mode colors
local function set_terminal_colors(colors)
  vim.g.terminal_color_0 = colors.bg0
  vim.g.terminal_color_8 = colors.gray
  vim.g.terminal_color_1 = colors.neutral_red
  vim.g.terminal_color_9 = colors.red
  vim.g.terminal_color_2 = colors.neutral_green
  vim.g.terminal_color_10 = colors.green
  vim.g.terminal_color_3 = colors.neutral_yellow
  vim.g.terminal_color_11 = colors.yellow
  vim.g.terminal_color_4 = colors.neutral_blue
  vim.g.terminal_color_12 = colors.blue
  vim.g.terminal_color_5 = colors.neutral_purple
  vim.g.terminal_color_13 = colors.purple
  vim.g.terminal_color_6 = colors.neutral_aqua
  vim.g.terminal_color_14 = colors.aqua
  vim.g.terminal_color_7 = colors.fg4
  vim.g.terminal_color_15 = colors.fg1
end

groups.setup = function()
  local config = require("gruvbox").config

  local palette = require("gruvbox.palette")
  for k, v in pairs(config.palette_overrides) do
    palette[k] = v
  end

  local colors = get_base_colors(palette, config.contrast)
  set_terminal_colors(colors)

  local groups = {
    -- Base groups
    GruvboxFg0 = { fg = colors.fg0 },
    GruvboxFg1 = { fg = colors.fg1 },
    GruvboxFg2 = { fg = colors.fg2 },
    GruvboxFg3 = { fg = colors.fg3 },
    GruvboxFg4 = { fg = colors.fg4 },
    GruvboxGray = { fg = colors.gray },
    GruvboxBg0 = { fg = colors.bg0 },
    GruvboxBg1 = { fg = colors.bg1 },
    GruvboxBg2 = { fg = colors.bg2 },
    GruvboxBg3 = { fg = colors.bg3 },
    GruvboxBg4 = { fg = colors.bg4 },
    GruvboxRed = { fg = colors.red },
    GruvboxRedBold = { fg = colors.red, bold = config.bold },
    GruvboxGreen = { fg = colors.green },
    GruvboxGreenBold = { fg = colors.green, bold = config.bold },
    GruvboxYellow = { fg = colors.yellow },
    GruvboxYellowBold = { fg = colors.yellow, bold = config.bold },
    GruvboxBlue = { fg = colors.blue },
    GruvboxBlueBold = { fg = colors.blue, bold = config.bold },
    GruvboxPurple = { fg = colors.purple },
    GruvboxPurpleBold = { fg = colors.purple, bold = config.bold },
    GruvboxAqua = { fg = colors.aqua },
    GruvboxAquaBold = { fg = colors.aqua, bold = config.bold },
    GruvboxOrange = { fg = colors.orange },
    GruvboxOrangeBold = { fg = colors.orange, bold = config.bold },
    GruvboxRedSign = { fg = colors.red, bg = colors.bg1, reverse = config.invert_signs },
    GruvboxGreenSign = { fg = colors.green, bg = colors.bg1, reverse = config.invert_signs },
    GruvboxYellowSign = { fg = colors.yellow, bg = colors.bg1, reverse = config.invert_signs },
    GruvboxBlueSign = { fg = colors.blue, bg = colors.bg1, reverse = config.invert_signs },
    GruvboxPurpleSign = { fg = colors.purple, bg = colors.bg1, reverse = config.invert_signs },
    GruvboxAquaSign = { fg = colors.aqua, bg = colors.bg1, reverse = config.invert_signs },
    GruvboxOrangeSign = { fg = colors.orange, bg = colors.bg1, reverse = config.invert_signs },
    GruvboxRedUnderline = { undercurl = config.undercurl, sp = colors.red },
    GruvboxGreenUnderline = { undercurl = config.undercurl, sp = colors.green },
    GruvboxYellowUnderline = { undercurl = config.undercurl, sp = colors.yellow },
    GruvboxBlueUnderline = { undercurl = config.undercurl, sp = colors.blue },
    GruvboxPurpleUnderline = { undercurl = config.undercurl, sp = colors.purple },
    GruvboxAquaUnderline = { undercurl = config.undercurl, sp = colors.aqua },
    GruvboxOrangeUnderline = { undercurl = config.undercurl, sp = colors.orange },
    Normal = config.transparent_mode and { fg = colors.fg1, bg = nil } or { fg = colors.fg1, bg = colors.bg0 },
    NormalNC = config.dim_inactive and { fg = colors.fg0, bg = colors.bg1 } or { link = "Normal" },
    CursorLine = { bg = colors.bg1 },
    CursorColumn = { link = "CursorLine" },
    TabLineFill = { fg = colors.bg4, bg = colors.bg1, reverse = config.invert_tabline },
    TabLineSel = { fg = colors.green, bg = colors.bg1, reverse = config.invert_tabline },
    TabLine = { link = "TabLineFill" },
    MatchParen = { bg = colors.bg3, bold = config.bold },
    ColorColumn = { bg = colors.bg1 },
    Conceal = { fg = colors.blue },
    CursorLineNr = { fg = colors.yellow, bg = colors.bg1 },
    NonText = { link = "GruvboxBg2" },
    SpecialKey = { link = "GruvboxFg4" },
    Visual = { bg = colors.bg3, reverse = config.invert_selection },
    VisualNOS = { link = "Visual" },
    Search = { fg = colors.yellow, bg = colors.bg0, reverse = config.inverse },
    IncSearch = { fg = colors.orange, bg = colors.bg0, reverse = config.inverse },
    CurSearch = { link = "IncSearch" },
    QuickFixLine = { fg = colors.bg0, bg = colors.yellow, bold = config.bold },
    Underlined = { fg = colors.blue, underline = config.underline },
    StatusLine = { fg = colors.bg2, bg = colors.fg1, reverse = config.inverse },
    StatusLineNC = { fg = colors.bg1, bg = colors.fg4, reverse = config.inverse },
    WinBar = { fg = colors.fg4, bg = colors.bg0 },
    WinBarNC = { fg = colors.fg3, bg = colors.bg1 },
    VertSplit = { fg = colors.bg3, bg = colors.bg0 },
    WildMenu = { fg = colors.blue, bg = colors.bg2, bold = config.bold },
    Directory = { link = "GruvboxBlueBold" },
    Title = { link = "GruvboxGreenBold" },
    ErrorMsg = { fg = colors.bg0, bg = colors.red, bold = config.bold },
    MoreMsg = { link = "GruvboxYellowBold" },
    ModeMsg = { link = "GruvboxYellowBold" },
    Question = { link = "GruvboxOrangeBold" },
    WarningMsg = { link = "GruvboxRedBold" },
    LineNr = { fg = colors.bg4 },
    SignColumn = config.transparent_mode and { bg = nil } or { bg = colors.bg1 },
    Folded = { fg = colors.gray, bg = colors.bg1, italic = config.italic },
    FoldColumn = { fg = colors.gray, bg = colors.bg1 },
    Cursor = { reverse = config.inverse },
    vCursor = { link = "Cursor" },
    iCursor = { link = "Cursor" },
    lCursor = { link = "Cursor" },
    Special = { link = "GruvboxOrange" },
    Comment = { fg = colors.gray, italic = config.italic },
    Todo = { fg = colors.fg0, bold = config.bold, italic = config.italic },
    Done = { fg = colors.orange, bold = config.bold, italic = config.italic },
    Error = { fg = colors.red, bold = config.bold, reverse = config.inverse },
    Statement = { link = "GruvboxRed" },
    Conditional = { link = "GruvboxRed" },
    Repeat = { link = "GruvboxRed" },
    Label = { link = "GruvboxRed" },
    Exception = { link = "GruvboxRed" },
    Operator = { fg = colors.orange, italic = config.italic },
    Keyword = { link = "GruvboxRed" },
    Identifier = { link = "GruvboxBlue" },
    Function = { link = "GruvboxGreenBold" },
    PreProc = { link = "GruvboxAqua" },
    Include = { link = "GruvboxAqua" },
    Define = { link = "GruvboxAqua" },
    Macro = { link = "GruvboxAqua" },
    PreCondit = { link = "GruvboxAqua" },
    Constant = { link = "GruvboxPurple" },
    Character = { link = "GruvboxPurple" },
    String = { fg = colors.green, italic = config.italic },
    Boolean = { link = "GruvboxPurple" },
    Number = { link = "GruvboxPurple" },
    Float = { link = "GruvboxPurple" },
    Type = { link = "GruvboxYellow" },
    StorageClass = { link = "GruvboxOrange" },
    Structure = { link = "GruvboxAqua" },
    Typedef = { link = "GruvboxYellow" },
    Pmenu = { fg = colors.fg1, bg = colors.bg2 },
    PmenuSel = { fg = colors.bg2, bg = colors.blue, bold = config.bold },
    PmenuSbar = { bg = colors.bg2 },
    PmenuThumb = { bg = colors.bg4 },
    DiffDelete = { fg = colors.red, bg = colors.bg0, reverse = config.inverse },
    DiffAdd = { fg = colors.green, bg = colors.bg0, reverse = config.inverse },
    DiffChange = { fg = colors.aqua, bg = colors.bg0, reverse = config.inverse },
    DiffText = { fg = colors.yellow, bg = colors.bg0, reverse = config.inverse },
    SpellCap = { link = "GruvboxBlueUnderline" },
    SpellBad = { link = "GruvboxRedUnderline" },
    SpellLocal = { link = "GruvboxAquaUnderline" },
    SpellRare = { link = "GruvboxPurpleUnderline" },
    -- LSP Diagnostic
    DiagnosticError = { link = "GruvboxRed" },
    DiagnosticSignError = { link = "GruvboxRedSign" },
    DiagnosticUnderlineError = { link = "GruvboxRedUnderline" },
    DiagnosticWarn = { link = "GruvboxYellow" },
    DiagnosticSignWarn = { link = "GruvboxYellowSign" },
    DiagnosticUnderlineWarn = { link = "GruvboxYellowUnderline" },
    DiagnosticInfo = { link = "GruvboxBlue" },
    DiagnosticSignInfo = { link = "GruvboxBlueSign" },
    DiagnosticUnderlineInfo = { link = "GruvboxBlueUnderline" },
    DiagnosticHint = { link = "GruvboxAqua" },
    DiagnosticSignHint = { link = "GruvboxAquaSign" },
    DiagnosticUnderlineHint = { link = "GruvboxAquaUnderline" },
    DiagnosticFloatingError = { link = "GruvboxRed" },
    DiagnosticFloatingWarn = { link = "GruvboxOrange" },
    DiagnosticFloatingInfo = { link = "GruvboxBlue" },
    DiagnosticFloatingHint = { link = "GruvboxAqua" },
    DiagnosticVirtualTextError = { link = "GruvboxRed" },
    DiagnosticVirtualTextWarn = { link = "GruvboxYellow" },
    DiagnosticVirtualTextInfo = { link = "GruvboxBlue" },
    DiagnosticVirtualTextHint = { link = "GruvboxAqua" },
    LspReferenceRead = { link = "GruvboxYellowBold" },
    LspReferenceText = { link = "GruvboxYellowBold" },
    LspReferenceWrite = { link = "GruvboxOrangeBold" },
    LspCodeLens = { link = "GruvboxGray" },
    -- nvim-treesitter (0.8 compat)
    -- Adapted from https://github.com/nvim-treesitter/nvim-treesitter/commit/42ab95d5e11f247c6f0c8f5181b02e816caa4a4f#commitcomment-87014462
    ["@comment"] = { link = "Comment" },
    ["@none"] = { bg = "NONE", fg = "NONE" },
    ["@preproc"] = { link = "PreProc" },
    ["@define"] = { link = "Define" },
    ["@operator"] = { link = "Operator" },
    ["@punctuation.delimiter"] = { link = "Delimiter" },
    ["@punctuation.bracket"] = { link = "Delimiter" },
    ["@punctuation.special"] = { link = "Delimiter" },
    ["@string"] = { link = "String" },
    ["@string.regex"] = { link = "String" },
    ["@string.escape"] = { link = "SpecialChar" },
    ["@string.special"] = { link = "SpecialChar" },
    ["@character"] = { link = "Character" },
    ["@character.special"] = { link = "SpecialChar" },
    ["@boolean"] = { link = "Boolean" },
    ["@number"] = { link = "Number" },
    ["@float"] = { link = "Float" },
    ["@function"] = { link = "Function" },
    ["@function.call"] = { link = "Function" },
    ["@function.builtin"] = { link = "Special" },
    ["@function.macro"] = { link = "Macro" },
    ["@method"] = { link = "Function" },
    ["@method.call"] = { link = "Function" },
    ["@constructor"] = { link = "Special" },
    ["@parameter"] = { link = "Identifier" },
    ["@keyword"] = { link = "Keyword" },
    ["@keyword.function"] = { link = "Keyword" },
    ["@keyword.return"] = { link = "Keyword" },
    ["@conditional"] = { link = "Conditional" },
    ["@repeat"] = { link = "Repeat" },
    ["@debug"] = { link = "Debug" },
    ["@label"] = { link = "Label" },
    ["@include"] = { link = "Include" },
    ["@exception"] = { link = "Exception" },
    ["@type"] = { link = "Type" },
    ["@type.builtin"] = { link = "Type" },
    ["@type.qualifier"] = { link = "Type" },
    ["@type.definition"] = { link = "Typedef" },
    ["@storageclass"] = { link = "StorageClass" },
    ["@attribute"] = { link = "PreProc" },
    ["@field"] = { link = "Identifier" },
    ["@property"] = { link = "Identifier" },
    ["@variable"] = { link = "GruvboxFg1" },
    ["@variable.builtin"] = { link = "Special" },
    ["@constant"] = { link = "Constant" },
    ["@constant.builtin"] = { link = "Special" },
    ["@constant.macro"] = { link = "Define" },
    ["@namespace"] = { link = "Include" },
    ["@symbol"] = { link = "Identifier" },
    ["@text"] = { link = "GruvboxFg1" },
    ["@text.title"] = { link = "Title" },
    ["@text.literal"] = { link = "String" },
    ["@text.uri"] = { link = "Underlined" },
    ["@text.math"] = { link = "Special" },
    ["@text.environment"] = { link = "Macro" },
    ["@text.environment.name"] = { link = "Type" },
    ["@text.reference"] = { link = "Constant" },
    ["@text.todo"] = { link = "Todo" },
    ["@text.todo.unchecked"] = { link = "Todo" },
    ["@text.todo.checked"] = { link = "Done" },
    ["@text.note"] = { link = "SpecialComment" },
    ["@text.warning"] = { link = "WarningMsg" },
    ["@text.danger"] = { link = "ErrorMsg" },
    ["@text.diff.add"] = { link = "diffAdded" },
    ["@text.diff.delete"] = { link = "diffRemoved" },
    ["@tag"] = { link = "Tag" },
    ["@tag.attribute"] = { link = "Identifier" },
    ["@tag.delimiter"] = { link = "Delimiter" },

    -- nvim-treesitter (0.8 overrides)
    ["@text.strong"] = { bold = config.bold },
    ["@text.strike"] = { strikethrough = config.strikethrough },
    ["@text.emphasis"] = { italic = config.italic },
    ["@text.underline"] = { underline = config.underline },
    ["@keyword.operator"] = { link = "GruvboxRed" },
    -- gitcommit
    gitcommitSelectedFile = { link = "GruvboxGreen" },
    gitcommitDiscardedFile = { link = "GruvboxRed" },
    -- gitsigns.nvim
    GitSignsAdd = { link = "GruvboxGreenSign" },
    GitSignsChange = { link = "GruvboxAquaSign" },
    GitSignsDelete = { link = "GruvboxRedSign" },
    -- nvim-tree
    NvimTreeSymlink = { fg = palette.neutral_aqua },
    NvimTreeRootFolder = { fg = palette.neutral_purple, bold = true },
    NvimTreeFolderIcon = { fg = palette.neutral_blue, bold = true },
    NvimTreeFileIcon = { fg = palette.light2 },
    NvimTreeExecFile = { fg = palette.neutral_green, bold = true },
    NvimTreeOpenedFile = { fg = palette.bright_red, bold = true },
    NvimTreeSpecialFile = { fg = palette.neutral_yellow, bold = true, underline = true },
    NvimTreeImageFile = { fg = palette.neutral_purple },
    NvimTreeIndentMarker = { fg = palette.dark3 },
    NvimTreeGitDirty = { fg = palette.neutral_yellow },
    NvimTreeGitStaged = { fg = palette.neutral_yellow },
    NvimTreeGitMerge = { fg = palette.neutral_purple },
    NvimTreeGitRenamed = { fg = palette.neutral_purple },
    NvimTreeGitNew = { fg = palette.neutral_yellow },
    NvimTreeGitDeleted = { fg = palette.neutral_red },
    NvimTreeWindowPicker = { bg = palette.faded_aqua },
    -- termdebug
    debugPC = { bg = palette.faded_blue },
    debugBreakpoint = { link = "GruvboxRedSign" },
    -- vim-startify
    StartifyBracket = { link = "GruvboxFg3" },
    StartifyFile = { link = "GruvboxFg1" },
    StartifyNumber = { link = "GruvboxBlue" },
    StartifyPath = { link = "GruvboxGray" },
    StartifySlash = { link = "GruvboxGray" },
    StartifySection = { link = "GruvboxYellow" },
    StartifySpecial = { link = "GruvboxBg2" },
    StartifyHeader = { link = "GruvboxOrange" },
    StartifyFooter = { link = "GruvboxBg2" },
    StartifyVar = { link = "StartifyPath" },
    StartifySelect = { link = "Title" },
    -- vim-dirvish
    DirvishPathTail = { link = "GruvboxAqua" },
    DirvishArg = { link = "GruvboxYellow" },
    -- netrw
    netrwDir = { link = "GruvboxAqua" },
    netrwClassify = { link = "GruvboxAqua" },
    netrwLink = { link = "GruvboxGray" },
    netrwSymLink = { link = "GruvboxFg1" },
    netrwExe = { link = "GruvboxYellow" },
    netrwComment = { link = "GruvboxGray" },
    netrwList = { link = "GruvboxBlue" },
    netrwHelpCmd = { link = "GruvboxAqua" },
    netrwCmdSep = { link = "GruvboxFg3" },
    netrwVersion = { link = "GruvboxGreen" },
    -- nerdtree
    NERDTreeDir = { link = "GruvboxAqua" },
    NERDTreeDirSlash = { link = "GruvboxAqua" },
    NERDTreeOpenable = { link = "GruvboxOrange" },
    NERDTreeClosable = { link = "GruvboxOrange" },
    NERDTreeFile = { link = "GruvboxFg1" },
    NERDTreeExecFile = { link = "GruvboxYellow" },
    NERDTreeUp = { link = "GruvboxGray" },
    NERDTreeCWD = { link = "GruvboxGreen" },
    NERDTreeHelp = { link = "GruvboxFg1" },
    NERDTreeToggleOn = { link = "GruvboxGreen" },
    NERDTreeToggleOff = { link = "GruvboxRed" },
    -- coc.nvim
    CocErrorSign = { link = "GruvboxRedSign" },
    CocWarningSign = { link = "GruvboxOrangeSign" },
    CocInfoSign = { link = "GruvboxBlueSign" },
    CocHintSign = { link = "GruvboxAquaSign" },
    CocErrorFloat = { link = "GruvboxRed" },
    CocWarningFloat = { link = "GruvboxOrange" },
    CocInfoFloat = { link = "GruvboxBlue" },
    CocHintFloat = { link = "GruvboxAqua" },
    CocDiagnosticsError = { link = "GruvboxRed" },
    CocDiagnosticsWarning = { link = "GruvboxOrange" },
    CocDiagnosticsInfo = { link = "GruvboxBlue" },
    CocDiagnosticsHint = { link = "GruvboxAqua" },
    CocSelectedText = { link = "GruvboxRed" },
    CocMenuSel = { link = "PmenuSel" },
    CocCodeLens = { link = "GruvboxGray" },
    CocErrorHighlight = { link = "GruvboxRedUnderline" },
    CocWarningHighlight = { link = "GruvboxOrangeUnderline" },
    CocInfoHighlight = { link = "GruvboxBlueUnderline" },
    CocHintHighlight = { link = "GruvboxAquaUnderline" },
    -- telescope.nvim
    TelescopeNormal = { link = "GruvboxFg1" },
    TelescopeSelection = { link = "GruvboxOrangeBold" },
    TelescopeSelectionCaret = { link = "GruvboxRed" },
    TelescopeMultiSelection = { link = "GruvboxGray" },
    TelescopeBorder = { link = "TelescopeNormal" },
    TelescopePromptBorder = { link = "TelescopeNormal" },
    TelescopeResultsBorder = { link = "TelescopeNormal" },
    TelescopePreviewBorder = { link = "TelescopeNormal" },
    TelescopeMatching = { link = "GruvboxBlue" },
    TelescopePromptPrefix = { link = "GruvboxRed" },
    TelescopePrompt = { link = "TelescopeNormal" },
    -- nvim-cmp
    CmpItemAbbr = { link = "GruvboxFg0" },
    CmpItemAbbrDeprecated = { link = "GruvboxFg1" },
    CmpItemAbbrMatch = { link = "GruvboxBlueBold" },
    CmpItemAbbrMatchFuzzy = { link = "GruvboxBlueUnderline" },
    CmpItemMenu = { link = "GruvboxGray" },
    CmpItemKindText = { link = "GruvboxOrange" },
    CmpItemKindMethod = { link = "GruvboxBlue" },
    CmpItemKindFunction = { link = "GruvboxBlue" },
    CmpItemKindConstructor = { link = "GruvboxYellow" },
    CmpItemKindField = { link = "GruvboxBlue" },
    CmpItemKindClass = { link = "GruvboxYellow" },
    CmpItemKindInterface = { link = "GruvboxYellow" },
    CmpItemKindModule = { link = "GruvboxBlue" },
    CmpItemKindProperty = { link = "GruvboxBlue" },
    CmpItemKindValue = { link = "GruvboxOrange" },
    CmpItemKindEnum = { link = "GruvboxYellow" },
    CmpItemKindKeyword = { link = "GruvboxPurple" },
    CmpItemKindSnippet = { link = "GruvboxGreen" },
    CmpItemKindFile = { link = "GruvboxBlue" },
    CmpItemKindEnumMember = { link = "GruvboxAqua" },
    CmpItemKindConstant = { link = "GruvboxOrange" },
    CmpItemKindStruct = { link = "GruvboxYellow" },
    CmpItemKindTypeParameter = { link = "GruvboxYellow" },
    diffAdded = { link = "GruvboxGreen" },
    diffRemoved = { link = "GruvboxRed" },
    diffChanged = { link = "GruvboxAqua" },
    diffFile = { link = "GruvboxOrange" },
    diffNewFile = { link = "GruvboxYellow" },
    diffLine = { link = "GruvboxBlue" },
    -- navic (highlight icons)
    NavicIconsFile = { link = "GruvboxBlue" },
    NavicIconsModule = { link = "GruvboxOrange" },
    NavicIconsNamespace = { link = "GruvboxBlue" },
    NavicIconsPackage = { link = "GruvboxAqua" },
    NavicIconsClass = { link = "GruvboxYellow" },
    NavicIconsMethod = { link = "GruvboxBlue" },
    NavicIconsProperty = { link = "GruvboxAqua" },
    NavicIconsField = { link = "GruvboxPurple" },
    NavicIconsConstructor = { link = "GruvboxBlue" },
    NavicIconsEnum = { link = "GruvboxPurple" },
    NavicIconsInterface = { link = "GruvboxGreen" },
    NavicIconsFunction = { link = "GruvboxBlue" },
    NavicIconsVariable = { link = "GruvboxPurple" },
    NavicIconsConstant = { link = "GruvboxOrange" },
    NavicIconsString = { link = "GruvboxGreen" },
    NavicIconsNumber = { link = "GruvboxOrange" },
    NavicIconsBoolean = { link = "GruvboxOrange" },
    NavicIconsArray = { link = "GruvboxOrange" },
    NavicIconsObject = { link = "GruvboxOrange" },
    NavicIconsKey = { link = "GruvboxAqua" },
    NavicIconsNull = { link = "GruvboxOrange" },
    NavicIconsEnumMember = { link = "GruvboxYellow" },
    NavicIconsStruct = { link = "GruvboxPurple" },
    NavicIconsEvent = { link = "GruvboxYellow" },
    NavicIconsOperator = { link = "GruvboxRed" },
    NavicIconsTypeParameter = { link = "GruvboxRed" },
    NavicText = { link = "GruvboxWhite" },
    NavicSeparator = { link = "GruvboxWhite" },
    -- html
    htmlTag = { link = "GruvboxAquaBold" },
    htmlEndTag = { link = "GruvboxAquaBold" },
    htmlTagName = { link = "GruvboxBlue" },
    htmlArg = { link = "GruvboxOrange" },
    htmlTagN = { link = "GruvboxFg1" },
    htmlSpecialTagName = { link = "GruvboxBlue" },
    htmlLink = { fg = colors.fg4, underline = config.underline },
    htmlSpecialChar = { link = "GruvboxRed" },
    htmlBold = { fg = colors.fg0, bg = colors.bg0, bold = config.bold },
    htmlBoldUnderline = { fg = colors.fg0, bg = colors.bg0, bold = config.bold, underline = config.underline },
    htmlBoldItalic = { fg = colors.fg0, bg = colors.bg0, bold = config.bold, italic = config.italic },
    htmlBoldUnderlineItalic = {
      fg = colors.fg0,
      bg = colors.bg0,
      bold = config.bold,
      italic = config.italic,
      underline = config.underline,
    },
    htmlUnderline = { fg = colors.fg0, bg = colors.bg0, underline = config.underline },
    htmlUnderlineItalic = {
      fg = colors.fg0,
      bg = colors.bg0,
      italic = config.italic,
      underline = config.underline,
    },
    htmlItalic = { fg = colors.fg0, bg = colors.bg0, bold = config.italic },
    -- xml
    xmlTag = { link = "GruvboxAquaBold" },
    xmlEndTag = { link = "GruvboxAquaBold" },
    xmlTagName = { link = "GruvboxBlue" },
    xmlEqual = { link = "GruvboxBlue" },
    docbkKeyword = { link = "GruvboxAquaBold" },
    xmlDocTypeDecl = { link = "GruvboxGray" },
    xmlDocTypeKeyword = { link = "GruvboxPurple" },
    xmlCdataStart = { link = "GruvboxGray" },
    xmlCdataCdata = { link = "GruvboxPurple" },
    dtdFunction = { link = "GruvboxGray" },
    dtdTagName = { link = "GruvboxPurple" },
    xmlAttrib = { link = "GruvboxOrange" },
    xmlProcessingDelim = { link = "GruvboxGray" },
    dtdParamEntityPunct = { link = "GruvboxGray" },
    dtdParamEntityDPunct = { link = "GruvboxGray" },
    xmlAttribPunct = { link = "GruvboxGray" },
    xmlEntity = { link = "GruvboxRed" },
    xmlEntityPunct = { link = "GruvboxRed" },
    -- clojure
    clojureKeyword = { link = "GruvboxBlue" },
    clojureCond = { link = "GruvboxOrange" },
    clojureSpecial = { link = "GruvboxOrange" },
    clojureDefine = { link = "GruvboxOrange" },
    clojureFunc = { link = "GruvboxYellow" },
    clojureRepeat = { link = "GruvboxYellow" },
    clojureCharacter = { link = "GruvboxAqua" },
    clojureStringEscape = { link = "GruvboxAqua" },
    clojureException = { link = "GruvboxRed" },
    clojureRegexp = { link = "GruvboxAqua" },
    clojureRegexpEscape = { link = "GruvboxAqua" },
    clojureRegexpCharClass = { fg = colors.fg3, bold = config.bold },
    clojureRegexpMod = { link = "clojureRegexpCharClass" },
    clojureRegexpQuantifier = { link = "clojureRegexpCharClass" },
    clojureParen = { link = "GruvboxFg3" },
    clojureAnonArg = { link = "GruvboxYellow" },
    clojureVariable = { link = "GruvboxBlue" },
    clojureMacro = { link = "GruvboxOrange" },
    clojureMeta = { link = "GruvboxYellow" },
    clojureDeref = { link = "GruvboxYellow" },
    clojureQuote = { link = "GruvboxYellow" },
    clojureUnquote = { link = "GruvboxYellow" },
    -- C
    cOperator = { link = "GruvboxPurple" },
    cppOperator = { link = "GruvboxPurple" },
    cStructure = { link = "GruvboxOrange" },
    -- python
    pythonBuiltin = { link = "GruvboxOrange" },
    pythonBuiltinObj = { link = "GruvboxOrange" },
    pythonBuiltinFunc = { link = "GruvboxOrange" },
    pythonFunction = { link = "GruvboxAqua" },
    pythonDecorator = { link = "GruvboxRed" },
    pythonInclude = { link = "GruvboxBlue" },
    pythonImport = { link = "GruvboxBlue" },
    pythonRun = { link = "GruvboxBlue" },
    pythonCoding = { link = "GruvboxBlue" },
    pythonOperator = { link = "GruvboxRed" },
    pythonException = { link = "GruvboxRed" },
    pythonExceptions = { link = "GruvboxPurple" },
    pythonBoolean = { link = "GruvboxPurple" },
    pythonDot = { link = "GruvboxFg3" },
    pythonConditional = { link = "GruvboxRed" },
    pythonRepeat = { link = "GruvboxRed" },
    pythonDottedName = { link = "GruvboxGreenBold" },
    -- CSS
    cssBraces = { link = "GruvboxBlue" },
    cssFunctionName = { link = "GruvboxYellow" },
    cssIdentifier = { link = "GruvboxOrange" },
    cssClassName = { link = "GruvboxGreen" },
    cssColor = { link = "GruvboxBlue" },
    cssSelectorOp = { link = "GruvboxBlue" },
    cssSelectorOp2 = { link = "GruvboxBlue" },
    cssImportant = { link = "GruvboxGreen" },
    cssVendor = { link = "GruvboxFg1" },
    cssTextProp = { link = "GruvboxAqua" },
    cssAnimationProp = { link = "GruvboxAqua" },
    cssUIProp = { link = "GruvboxYellow" },
    cssTransformProp = { link = "GruvboxAqua" },
    cssTransitionProp = { link = "GruvboxAqua" },
    cssPrintProp = { link = "GruvboxAqua" },
    cssPositioningProp = { link = "GruvboxYellow" },
    cssBoxProp = { link = "GruvboxAqua" },
    cssFontDescriptorProp = { link = "GruvboxAqua" },
    cssFlexibleBoxProp = { link = "GruvboxAqua" },
    cssBorderOutlineProp = { link = "GruvboxAqua" },
    cssBackgroundProp = { link = "GruvboxAqua" },
    cssMarginProp = { link = "GruvboxAqua" },
    cssListProp = { link = "GruvboxAqua" },
    cssTableProp = { link = "GruvboxAqua" },
    cssFontProp = { link = "GruvboxAqua" },
    cssPaddingProp = { link = "GruvboxAqua" },
    cssDimensionProp = { link = "GruvboxAqua" },
    cssRenderProp = { link = "GruvboxAqua" },
    cssColorProp = { link = "GruvboxAqua" },
    cssGeneratedContentProp = { link = "GruvboxAqua" },
    -- javascript
    javaScriptBraces = { link = "GruvboxFg1" },
    javaScriptFunction = { link = "GruvboxAqua" },
    javaScriptIdentifier = { link = "GruvboxRed" },
    javaScriptMember = { link = "GruvboxBlue" },
    javaScriptNumber = { link = "GruvboxPurple" },
    javaScriptNull = { link = "GruvboxPurple" },
    javaScriptParens = { link = "GruvboxFg3" },
    -- typescript
    typescriptReserved = { link = "GruvboxAqua" },
    typescriptLabel = { link = "GruvboxAqua" },
    typescriptFuncKeyword = { link = "GruvboxAqua" },
    typescriptIdentifier = { link = "GruvboxOrange" },
    typescriptBraces = { link = "GruvboxFg1" },
    typescriptEndColons = { link = "GruvboxFg1" },
    typescriptDOMObjects = { link = "GruvboxFg1" },
    typescriptAjaxMethods = { link = "GruvboxFg1" },
    typescriptLogicSymbols = { link = "GruvboxFg1" },
    typescriptDocSeeTag = { link = "Comment" },
    typescriptDocParam = { link = "Comment" },
    typescriptDocTags = { link = "vimCommentTitle" },
    typescriptGlobalObjects = { link = "GruvboxFg1" },
    typescriptParens = { link = "GruvboxFg3" },
    typescriptOpSymbols = { link = "GruvboxFg3" },
    typescriptHtmlElemProperties = { link = "GruvboxFg1" },
    typescriptNull = { link = "GruvboxPurple" },
    typescriptInterpolationDelimiter = { link = "GruvboxAqua" },
    -- purescript
    purescriptModuleKeyword = { link = "GruvboxAqua" },
    purescriptModuleName = { link = "GruvboxFg1" },
    purescriptWhere = { link = "GruvboxAqua" },
    purescriptDelimiter = { link = "GruvboxFg4" },
    purescriptType = { link = "GruvboxFg1" },
    purescriptImportKeyword = { link = "GruvboxAqua" },
    purescriptHidingKeyword = { link = "GruvboxAqua" },
    purescriptAsKeyword = { link = "GruvboxAqua" },
    purescriptStructure = { link = "GruvboxAqua" },
    purescriptOperator = { link = "GruvboxBlue" },
    purescriptTypeVar = { link = "GruvboxFg1" },
    purescriptConstructor = { link = "GruvboxFg1" },
    purescriptFunction = { link = "GruvboxFg1" },
    purescriptConditional = { link = "GruvboxOrange" },
    purescriptBacktick = { link = "GruvboxOrange" },
    -- coffescript
    coffeeExtendedOp = { link = "GruvboxFg3" },
    coffeeSpecialOp = { link = "GruvboxFg3" },
    coffeeCurly = { link = "GruvboxOrange" },
    coffeeParen = { link = "GruvboxFg3" },
    coffeeBracket = { link = "GruvboxOrange" },
    -- ruby
    rubyStringDelimiter = { link = "GruvboxGreen" },
    rubyInterpolationDelimiter = { link = "GruvboxAqua" },
    rubyDefinedOperator = { link = "rubyKeyword" },
    -- objc
    objcTypeModifier = { link = "GruvboxRed" },
    objcDirective = { link = "GruvboxBlue" },
    -- go
    goDirective = { link = "GruvboxAqua" },
    goConstants = { link = "GruvboxPurple" },
    goDeclaration = { link = "GruvboxRed" },
    goDeclType = { link = "GruvboxBlue" },
    goBuiltins = { link = "GruvboxOrange" },
    -- lua
    luaIn = { link = "GruvboxRed" },
    luaFunction = { link = "GruvboxAqua" },
    luaTable = { link = "GruvboxOrange" },
    -- moonscript
    moonSpecialOp = { link = "GruvboxFg3" },
    moonExtendedOp = { link = "GruvboxFg3" },
    moonFunction = { link = "GruvboxFg3" },
    moonObject = { link = "GruvboxYellow" },
    -- java
    javaAnnotation = { link = "GruvboxBlue" },
    javaDocTags = { link = "GruvboxAqua" },
    javaCommentTitle = { link = "vimCommentTitle" },
    javaParen = { link = "GruvboxFg3" },
    javaParen1 = { link = "GruvboxFg3" },
    javaParen2 = { link = "GruvboxFg3" },
    javaParen3 = { link = "GruvboxFg3" },
    javaParen4 = { link = "GruvboxFg3" },
    javaParen5 = { link = "GruvboxFg3" },
    javaOperator = { link = "GruvboxOrange" },
    javaVarArg = { link = "GruvboxGreen" },
    -- elixir
    elixirDocString = { link = "Comment" },
    elixirStringDelimiter = { link = "GruvboxGreen" },
    elixirInterpolationDelimiter = { link = "GruvboxAqua" },
    elixirModuleDeclaration = { link = "GruvboxYellow" },
    -- scala
    scalaNameDefinition = { link = "GruvboxFg1" },
    scalaCaseFollowing = { link = "GruvboxFg1" },
    scalaCapitalWord = { link = "GruvboxFg1" },
    scalaTypeExtension = { link = "GruvboxFg1" },
    scalaKeyword = { link = "GruvboxRed" },
    scalaKeywordModifier = { link = "GruvboxRed" },
    scalaSpecial = { link = "GruvboxAqua" },
    scalaOperator = { link = "GruvboxFg1" },
    scalaTypeDeclaration = { link = "GruvboxYellow" },
    scalaTypeTypePostDeclaration = { link = "GruvboxYellow" },
    scalaInstanceDeclaration = { link = "GruvboxFg1" },
    scalaInterpolation = { link = "GruvboxAqua" },
    -- markdown
    markdownItalic = { fg = colors.fg3, italic = config.italic },
    markdownBold = { fg = colors.fg3, bold = config.bold },
    markdownBoldItalic = { fg = colors.fg3, bold = config.bold, italic = config.italic },
    markdownH1 = { link = "GruvboxGreenBold" },
    markdownH2 = { link = "GruvboxGreenBold" },
    markdownH3 = { link = "GruvboxYellowBold" },
    markdownH4 = { link = "GruvboxYellowBold" },
    markdownH5 = { link = "GruvboxYellow" },
    markdownH6 = { link = "GruvboxYellow" },
    markdownCode = { link = "GruvboxAqua" },
    markdownCodeBlock = { link = "GruvboxAqua" },
    markdownCodeDelimiter = { link = "GruvboxAqua" },
    markdownBlockquote = { link = "GruvboxGray" },
    markdownListMarker = { link = "GruvboxGray" },
    markdownOrderedListMarker = { link = "GruvboxGray" },
    markdownRule = { link = "GruvboxGray" },
    markdownHeadingRule = { link = "GruvboxGray" },
    markdownUrlDelimiter = { link = "GruvboxFg3" },
    markdownLinkDelimiter = { link = "GruvboxFg3" },
    markdownLinkTextDelimiter = { link = "GruvboxFg3" },
    markdownHeadingDelimiter = { link = "GruvboxOrange" },
    markdownUrl = { link = "GruvboxPurple" },
    markdownUrlTitleDelimiter = { link = "GruvboxGreen" },
    markdownLinkText = { fg = colors.gray, underline = config.underline },
    markdownIdDeclaration = { link = "markdownLinkText" },
    -- haskell
    haskellType = { link = "GruvboxBlue" },
    haskellIdentifier = { link = "GruvboxAqua" },
    haskellSeparator = { link = "GruvboxFg4" },
    haskellDelimiter = { link = "GruvboxOrange" },
    haskellOperators = { link = "GruvboxPurple" },
    haskellBacktick = { link = "GruvboxOrange" },
    haskellStatement = { link = "GruvboxPurple" },
    haskellConditional = { link = "GruvboxPurple" },
    haskellLet = { link = "GruvboxRed" },
    haskellDefault = { link = "GruvboxRed" },
    haskellWhere = { link = "GruvboxRed" },
    haskellBottom = { link = "GruvboxRedBold" },
    haskellImportKeywords = { link = "GruvboxPurpleBold" },
    haskellDeclKeyword = { link = "GruvboxOrange" },
    haskellDecl = { link = "GruvboxOrange" },
    haskellDeriving = { link = "GruvboxPurple" },
    haskellAssocType = { link = "GruvboxAqua" },
    haskellNumber = { link = "GruvboxAqua" },
    haskellPragma = { link = "GruvboxRedBold" },
    haskellTH = { link = "GruvboxAquaBold" },
    haskellForeignKeywords = { link = "GruvboxGreen" },
    haskellKeyword = { link = "GruvboxRed" },
    haskellFloat = { link = "GruvboxAqua" },
    haskellInfix = { link = "GruvboxPurple" },
    haskellQuote = { link = "GruvboxGreenBold" },
    haskellShebang = { link = "GruvboxYellowBold" },
    haskellLiquid = { link = "GruvboxPurpleBold" },
    haskellQuasiQuoted = { link = "GruvboxBlueBold" },
    haskellRecursiveDo = { link = "GruvboxPurple" },
    haskellQuotedType = { link = "GruvboxRed" },
    haskellPreProc = { link = "GruvboxFg4" },
    haskellTypeRoles = { link = "GruvboxRedBold" },
    haskellTypeForall = { link = "GruvboxRed" },
    haskellPatternKeyword = { link = "GruvboxBlue" },
    -- json
    jsonKeyword = { link = "GruvboxGreen" },
    jsonQuote = { link = "GruvboxGreen" },
    jsonBraces = { link = "GruvboxFg1" },
    jsonString = { link = "GruvboxFg1" },
    -- mail
    mailQuoted1 = { link = "GruvboxAqua" },
    mailQuoted2 = { link = "GruvboxPurple" },
    mailQuoted3 = { link = "GruvboxYellow" },
    mailQuoted4 = { link = "GruvboxGreen" },
    mailQuoted5 = { link = "GruvboxRed" },
    mailQuoted6 = { link = "GruvboxOrange" },
    mailSignature = { link = "Comment" },
    -- c#
    csBraces = { link = "GruvboxFg1" },
    csEndColon = { link = "GruvboxFg1" },
    csLogicSymbols = { link = "GruvboxFg1" },
    csParens = { link = "GruvboxFg3" },
    csOpSymbols = { link = "GruvboxFg3" },
    csInterpolationDelimiter = { link = "GruvboxFg3" },
    csInterpolationAlignDel = { link = "GruvboxAquaBold" },
    csInterpolationFormat = { link = "GruvboxAqua" },
    csInterpolationFormatDel = { link = "GruvboxAquaBold" },
    -- rust btw
    rustSigil = { link = "GruvboxOrange" },
    rustEscape = { link = "GruvboxAqua" },
    rustStringContinuation = { link = "GruvboxAqua" },
    rustEnum = { link = "GruvboxAqua" },
    rustStructure = { link = "GruvboxAqua" },
    rustModPathSep = { link = "GruvboxFg2" },
    rustCommentLineDoc = { link = "Comment" },
    rustDefault = { link = "GruvboxAqua" },
    -- ocaml
    ocamlOperator = { link = "GruvboxFg1" },
    ocamlKeyChar = { link = "GruvboxOrange" },
    ocamlArrow = { link = "GruvboxOrange" },
    ocamlInfixOpKeyword = { link = "GruvboxRed" },
    ocamlConstructor = { link = "GruvboxOrange" },
    -- lspsaga.nvim
    LspSagaCodeActionTitle = { link = "Title" },
    LspSagaCodeActionBorder = { link = "GruvboxFg1" },
    LspSagaCodeActionContent = { fg = colors.green, bold = config.bold },
    LspSagaLspFinderBorder = { link = "GruvboxFg1" },
    LspSagaAutoPreview = { link = "GruvboxOrange" },
    TargetWord = { fg = colors.blue, bold = config.bold },
    FinderSeparator = { link = "GruvboxAqua" },
    LspSagaDefPreviewBorder = { link = "GruvboxBlue" },
    LspSagaHoverBorder = { link = "GruvboxOrange" },
    LspSagaRenameBorder = { link = "GruvboxBlue" },
    LspSagaDiagnosticSource = { link = "GruvboxOrange" },
    LspSagaDiagnosticBorder = { link = "GruvboxPurple" },
    LspSagaDiagnosticHeader = { link = "GruvboxGreen" },
    LspSagaSignatureHelpBorder = { link = "GruvboxGreen" },
    SagaShadow = { link = "GruvboxBg0" },

    -- dashboard-nvim
    DashboardShortCut = { link = "GruvboxOrange" },
    DashboardHeader = { link = "GruvboxAqua" },
    DashboardCenter = { link = "GruvboxYellow" },
    DashboardFooter = { fg = colors.purple, italic = config.italic },
    -- mason
    MasonHighlight = { fg = palette.neutral_blue },
    MasonHighlightBlock = { fg = palette.dark0, bg = palette.neutral_blue },
    MasonHighlightBlockBold = { fg = palette.dark0, bg = palette.neutral_blue, bold = true },
    MasonHighlightSecondary = { fg = palette.neutral_yellow },
    MasonHighlightBlockSecondary = { fg = palette.dark0, bg = palette.neutral_yellow },
    MasonHighlightBlockBoldSecondary = { fg = palette.dark0, bg = palette.neutral_yellow, bold = true },
    MasonHeader = { link = "MasonHighlightBlockBoldSecondary" },
    MasonHeaderSecondary = { link = "MasonHighlightBlockBold" },
    MasonMuted = { fg = palette.light4 },
    MasonMutedBlock = { fg = palette.dark0, bg = palette.light4 },
    MasonMutedBlockBold = { fg = palette.dark0, bg = palette.light4, bold = true },
    -- lsp-inlayhints.nvim
    LspInlayHint = { link = "comment" },
    -- carbon.nvim
    CarbonFile = { link = "GruvboxFg1" },
    CarbonExe = { link = "GruvboxYellow" },
    CarbonSymlink = { link = "GruvboxAqua" },
    CarbonBrokenSymlink = { link = "GruvboxRed" },
    CarbonIndicator = { link = "GruvboxGray" },
    CarbonDanger = { link = "GruvboxRed" },
    CarbonPending = { link = "GruvboxYellow" },
  }

  for group, hl in pairs(config.overrides) do
    if groups[group] and not vim.tbl_isempty(hl) then
      groups[group].link = nil
    end
    groups[group] = vim.tbl_extend("force", groups[group] or {}, hl)
  end

  return groups
end

return groups
